"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol ? "symbol" : typeof obj; };

exports.default = function (_ref) {
  var t = _ref.types;

  return {
    visitor: {

      // process.env.NODE_ENV;
      MemberExpression: function MemberExpression(nodePath, state) {
        processNode(getReplacements(state.opts), nodePath, t.valueToNode, memberExpressionComparator);
      },


      // const x = { version: VERSION };
      Identifier: function Identifier(nodePath, state) {
        processNode(getReplacements(state.opts), nodePath, t.valueToNode, identifierComparator);
      },


      // typeof window
      UnaryExpression: function UnaryExpression(nodePath, state) {
        if (nodePath.node.operator !== "typeof") {
          return;
        }

        var replacements = getReplacements(state.opts);
        var keys = Object.keys(replacements);
        var typeofValues = {};

        keys.forEach(function (key) {
          if (key.substring(0, 7) === "typeof ") {
            typeofValues[key.substring(7)] = replacements[key];
          }
        });

        processNode(typeofValues, nodePath, t.valueToNode, unaryExpressionComparator);
      }
    }
  };
};

var fs = require("fs");
var path = require("path");

/**
 *  `babel-plugin-transfor-define` take options of two types: static config and a path to a file that
 *  can define config in any way a user sees fit. getReplacements takes the options and will either
 *  return the static config or get the dynamic config from disk
 * @param  {Object|String}  configOptions  configuration to parse
 * @return {Object}  replacement object
 */
var getReplacements = function getReplacements(configOptions) {
  if ((typeof configOptions === "undefined" ? "undefined" : _typeof(configOptions)) === "object") {
    return configOptions;
  }

  try {
    var fullPath = path.join(process.cwd(), configOptions);
    fs.accessSync(fullPath, fs.F_OK);
    return require(fullPath);
  } catch (err) {
    console.error("The nodePath: " + configOptions + " is not valid."); // eslint-disable-line
    throw new Error(err);
  }
};

/**
 * Replace a node with a given value. If the replacement results in a BinaryExpression, it will be
 * evaluated. For example, if the result of the replacement is `var x = "production" === "production"`
 * The evaluation will make a second replacement resulting in `var x = true`
 * @param  {function}   replaceFn    The function used to replace the node
 * @param  {babelNode}  nodePath     The node to evaluate
 * @param  {*}          replacement  The value the node will be replaced with
 * @return {undefined}
 */
var replaceAndEvaluateNode = function replaceAndEvaluateNode(replaceFn, nodePath, replacement) {
  nodePath.replaceWith(replaceFn(replacement));

  if (nodePath.parentPath.isBinaryExpression()) {
    var result = nodePath.parentPath.evaluate();

    if (result.confident) {
      nodePath.parentPath.replaceWith(replaceFn(result.value));
    }
  }
};

/**
 * Get the first value in the `obj` that causes the comparator to return true
 * @param  {Object}     obj         The object to search for replacements
 * @param  {function}   comparator  The function used to evaluate whether a node matches a value in `obj`
 * @param  {babelNode}  nodePath    The node to evaluate
 * @return {*}  The first matching value to replace OR null
 */
var getFirstReplacementValueForNode = function getFirstReplacementValueForNode(obj, comparator, nodePath) {
  var replacementKey = Object.keys(obj).filter(function (value) {
    return comparator(nodePath, value);
  }).shift();
  return obj[replacementKey] || null;
};

/**
 * Run the transformation over a node
 * @param  {Object}     replacements The object to search for replacements
 * @param  {babelNode}  nodePath     The node to evaluate
 * @param  {function}   replaceFn    The function used to replace the node
 * @param  {function}   comparator   The function used to evaluate whether a node matches a value in `replacements`
 * @return {undefined}
 */
var processNode = function processNode(replacements, nodePath, replaceFn, comparator) {
  // eslint-disable-line
  var replacement = getFirstReplacementValueForNode(replacements, comparator, nodePath);
  if (replacement) {
    replaceAndEvaluateNode(replaceFn, nodePath, replacement);
  }
};

var memberExpressionComparator = function memberExpressionComparator(nodePath, value) {
  return nodePath.matchesPattern(value);
};
var identifierComparator = function identifierComparator(nodePath, value) {
  return nodePath.node.name === value;
};
var unaryExpressionComparator = function unaryExpressionComparator(nodePath, value) {
  return nodePath.node.argument.name === value;
};